@@ -1,5 +1,5 @@
 /*
- *   stunnel       Universal SSL tunnel
+ *   stunnel       TLS offloading and load-balancing proxy
  *   Copyright (C) 1998-2015 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
@@ -65,7 +65,7 @@ NOEXPORT int main_unix(int argc, char* argv[]) {
     int fd;
 
     fd=open("/dev/null", O_RDWR); /* open /dev/null before chroot */
-    if(fd<0)
+    if(fd==INVALID_SOCKET)
         fatal("Could not open /dev/null");
 #endif
     main_init();
@@ -228,7 +228,7 @@ ICON_IMAGE load_icon_file(const char *file) {
 
 /**************************************** client callbacks */
 
-void ui_new_chain(const int section_number) {
+void ui_new_chain(const unsigned section_number) {
     (void)section_number; /* skip warning about unused parameter */
 }
 
